"""Models to represent slash commands and sub-commands."""
from __future__ import annotations

import inspect
import typing
from typing import Any, Optional, Type, Union

import docstring_parser
import nextcord
from docstring_parser.common import DocstringParam

from .options import ApplicationCommandOptionType, CommandOption
from .permissions import PermissionsSetter, warn_permissions_deprecation

if typing.TYPE_CHECKING:
    from .client import CommandClient


CommandCallback = Any
Permissions = Optional[Union[list["PermissionsSetter"], "PermissionsSetter"]]


class SlashCommandInvokeError(Exception):
    """An error raised when a slash command callback throws an error."""

    def __init__(self, original: Exception):
        """Store a reference to the original exception."""
        self.original = original
        super().__init__(
            f"Slash Command raised an exception: " f"{original.__class__.__name__}: {original}"
        )


class BaseSlashCommand:
    """Base class for slash commands and slash sub-commands."""

    name: str
    description: str

    def __init__(self, *, name: str, description: str):
        """Set up a new slash command."""
        self.name = name
        self.description = description

    def dump(self) -> dict[str, Any]:
        """Get the data to create the command."""
        data = {"name": self.name, "description": self.description}
        self._add_dump_data(data)
        return data

    def _add_dump_data(self, data: dict[str, Any]):
        """Add values to the data to create the command."""
        pass

    async def __call__(
        self,
        interaction: nextcord.Interaction,
        options: Optional[list[dict[str, Any]]] = None,
    ):
        """Handle the command or a nested subcommand of it being invoked."""
        if options:
            full_options = options
        else:
            full_options = interaction.data.get("options", []) if interaction.data else []
        option_map = {option["name"]: option for option in full_options}
        await self._process_option_data(interaction, option_map)  # type: ignore

    async def _process_option_data(
        self, interaction: nextcord.Interaction, options: dict[str, dict[str, Any]]
    ):
        """Process and use option data passed when the command is invoked."""
        raise NotImplementedError


class CallableSlashCommand(BaseSlashCommand):
    """Base class for slash commands that do something (not just groups)."""

    def __init__(self, *, callback: CommandCallback, name: str, description: str):
        """Set up the slash command."""
        BaseSlashCommand.__init__(self, name=name, description=description)
        self.callback = callback
        self.options: dict[str, CommandOption] = {}
        # Params to pass to the callback before the interaction. Useful for
        # 'self' / 'cls' parameters.
        self.prepend_params: tuple[Any, ...] = ()

    def _process_callback(self):
        """Process the name, docstring and arguments of the callback."""
        self.name = self.name or self.callback.__name__
        docstring = docstring_parser.parse(self.callback.__doc__ or "")
        if not self.description:
            self.description = docstring.short_description or "No description."
        self._process_options(docstring.params)

    def _process_options(self, descriptions: list[DocstringParam]):
        """Get the options for this command from callback annotations."""
        description_map = {
            docstring_param.arg_name: docstring_param.description
            for docstring_param in descriptions
        }
        signature = inspect.signature(self.callback)
        # get_type_hints will parse string type hints, which inspect won't.
        annotations = typing.get_type_hints(self.callback)
        done_self = False
        for n, parameter in enumerate(signature.parameters.values()):
            if n == 0 and parameter.name == "self":
                done_self = True
                continue
            if n == 0 or (done_self and n == 1):
                # It's the first parameter, the interaction.
                continue
            description = description_map.get(parameter.name)
            self._process_option(parameter, annotations[parameter.name], description)

    def _process_option(
        self, parameter: inspect.Parameter, annotation: Type, description: Optional[str]
    ):
        """Process an option annotation."""
        self.options[parameter.name] = CommandOption(
            name=parameter.name, description=description or "No description.", type=annotation
        )

    def _add_dump_data(self, data: dict[str, Any]):
        """Add the command options to the dump data."""
        data["options"] = data.get("options", [])
        for option in self.options.values():
            data["options"].append(option.dump())

    async def _process_option_data(
        self, interaction: nextcord.Interaction, options: dict[str, dict[str, Any]]
    ):
        """Process and use option data passed when the command is invoked."""
        arguments = {}
        for arg, option in self.options.items():
            data = options.get(option.name)
            arguments[arg] = await option(data, interaction)
        client: "CommandClient" = interaction._state._get_client()  # type: ignore
        try:
            if client.custom_interaction:
                interaction = client.custom_interaction(interaction)
            await self.callback(*self.prepend_params, interaction, **arguments)
        except Exception as exc:
            raise SlashCommandInvokeError(exc) from exc


class ContainerSlashCommand(BaseSlashCommand):
    """Base class for groups/commands which can have subcommands."""

    def __init__(self, name: str, description: str):
        """Set up the command group."""
        super().__init__(name=name, description=description)
        self.subcommands: dict[str, ChildSlashCommand] = {}

    def subcommand(
        self, name: Optional[str] = None, description: Optional[str] = None
    ) -> SlashSubCommandConstructor:
        """Create a decorator to register a new subcommand."""
        return SlashSubCommandConstructor(
            parent=self,
            name=name,
            description=description,
        )

    def _add_dump_data(self, data: dict[str, Any]):
        """Add the subcommands to the dump data."""
        data["options"] = data.get("options", [])
        for subcommand in self.subcommands.values():
            data["options"].append(subcommand.dump())

    async def _process_option_data(
        self, interaction: nextcord.Interaction, options: dict[str, dict[str, Any]]
    ):
        """Process and use option data passed when the command is invoked.

        Return value indicates whether the data was used.
        """
        for subcommand in self.subcommands.values():
            if subcommand.name in options:
                sub_options = options[subcommand.name].get("options", [])
                await subcommand(interaction, sub_options)
                return


class ChildSlashCommand(BaseSlashCommand):
    """Base class for groups/commands which must have parents."""


class TopLevelCommand(BaseSlashCommand):
    """Base class for groups/commands at the top level of the hierarchy.

    As far as the API is concerned this is all one type - application command,
    but the separation between a callable top level command and a container top
    level command helps for implementation.
    """

    def __init__(
        self,
        *,
        guild_id: Optional[int],
        # Permissions options remain only for legacy reasons, and have no effect.
        default_permission: bool | None = None,
        permissions: Permissions | None = None,
    ):
        """Set up a new top-level slash command."""
        if (default_permission is not None) or (permissions is not None):
            warn_permissions_deprecation()
        # We don't call the super constructor because this is a base class and
        # child classes will always have another parent with the same parent as
        # this class, which will call it instead.
        self.guild_id = guild_id
        self.id: Optional[int] = None


class SlashCommandGroup(TopLevelCommand, ContainerSlashCommand):
    """A top-level slash command that contains other commands."""

    def __init__(
        self,
        *,
        guild_id: Optional[int],
        name: str,
        description: str,
        # Permissions options remain only for legacy reasons, and have no effect.
        default_permission: bool | None = None,
        permissions: Permissions | None = None,
    ):
        """Set up a new top-level slash command."""
        if (default_permission is not None) or (permissions is not None):
            warn_permissions_deprecation()
        ContainerSlashCommand.__init__(self, name=name, description=description)
        TopLevelCommand.__init__(self, guild_id=guild_id)

    def _add_dump_data(self, data: dict[str, Any]):
        """Add the subcommands to the dump data."""
        ContainerSlashCommand._add_dump_data(self, data)


class SlashCommand(CallableSlashCommand, TopLevelCommand):
    """A callable top-level slash command."""

    def __init__(
        self,
        *,
        guild_id: Optional[int],
        callback: CommandCallback,
        name: str,
        description: str,
        # Permissions options remain only for legacy reasons, and have no effect.
        default_permission: bool | None = None,
        permissions: Permissions | None = None,
    ):
        """Set up a new top-level slash command."""
        if (default_permission is not None) or (permissions is not None):
            warn_permissions_deprecation()
        CallableSlashCommand.__init__(self, callback=callback, name=name, description=description)
        self._process_callback()
        TopLevelCommand.__init__(self, guild_id=guild_id)

    def _add_dump_data(self, data: dict[str, Any]):
        """Add the command options to the dump data."""
        CallableSlashCommand._add_dump_data(self, data)


class SlashCommandSubGroup(ChildSlashCommand, ContainerSlashCommand):
    """A sub-group of slash commands."""

    def __init__(self, name: str, description: str):
        """Set up the slash command group."""
        ContainerSlashCommand.__init__(self, name=name, description=description)

    def _add_dump_data(self, data: dict[str, Any]):
        """Add the option type to the dump data."""
        ContainerSlashCommand._add_dump_data(self, data)
        ChildSlashCommand._add_dump_data(self, data)
        data["type"] = ApplicationCommandOptionType.sub_command_group.value


class SlashSubCommand(ChildSlashCommand, CallableSlashCommand):
    """A subcommand of a slash command."""

    def __init__(self, callback: CommandCallback, name: str, description: str):
        """Set up a new subcommand."""
        CallableSlashCommand.__init__(self, callback=callback, name=name, description=description)
        self._process_callback()

    def _add_dump_data(self, data: dict[str, Any]):
        """Add the option type to the dump data."""
        CallableSlashCommand._add_dump_data(self, data)
        data["type"] = ApplicationCommandOptionType.sub_command.value


CT = typing.TypeVar("CT", bound=CallableSlashCommand)


class BaseSlashCommandConstructor(typing.Generic[CT]):
    """A class instances of which act as decorators to register commands."""

    command_class: Type[CT]

    def __init__(self, *, name: Optional[str], description: Optional[str]):
        """Set up the slash command constructor."""
        self.overwrites: dict[str, Any] = {"name": name, "description": description}

    def __call__(self, callback: CommandCallback) -> CT:
        """Construct an actual slash command."""
        command = self.command_class(callback=callback, **self.overwrites)
        self.register(command)
        return command

    def register(self, command: CT):
        """Register the created command."""
        raise NotImplementedError


class SlashCommandConstructor(BaseSlashCommandConstructor[SlashCommand]):
    """A class for decorating top-level commands."""

    command_class = SlashCommand

    def __init__(
        self,
        *,
        client: Optional["CommandClient"],
        guild_id: Optional[int],
        name: Optional[str],
        description: Optional[str],
        # Permissions options remain only for legacy reasons, and have no effect.
        default_permission: bool | None = None,
        permissions: Permissions | None = None,
    ):
        """Set up the slash command constructor."""
        if (default_permission is not None) or (permissions is not None):
            warn_permissions_deprecation()
        super().__init__(name=name, description=description)
        self.client = client
        self.overwrites["guild_id"] = guild_id

    def register(self, command: SlashCommand):
        """Register the created command."""
        if self.client:
            self.client._store_command(command)


class SlashSubCommandConstructor(BaseSlashCommandConstructor[SlashSubCommand]):
    """A class for decorating subcommands."""

    command_class = SlashSubCommand

    def __init__(
        self,
        *,
        parent: Optional[ContainerSlashCommand],
        name: Optional[str],
        description: Optional[str],
    ):
        """Set up the slash sub-command constructor."""
        super().__init__(name=name, description=description)
        self.parent = parent

    def register(self, command: SlashSubCommand):
        """Register the created subcommand."""
        if self.parent:
            self.parent.subcommands[command.name] = command



"""Definitions for the class-based command (sub)group system."""
from __future__ import annotations

import re
from typing import Any, Generic, Optional, TypeVar, Union

from .client import GUILD_ID_DEFAULT, GuildID
from .commands import (
    ChildSlashCommand,
    SlashCommandGroup,
    SlashCommandSubGroup,
    SlashSubCommand,
    SlashSubCommandConstructor,
)
from .permissions import warn_permissions_deprecation


def class_name_to_group_name(class_name: str) -> str:
    """Convert a class name to a group name."""
    return re.sub(r"([a-z])([A-Z0-9])", r"\1_\2", class_name).lower()


GT = TypeVar("GT", SlashCommandGroup, SlashCommandSubGroup)


class BaseCommandGroupMeta(type, Generic[GT]):
    """Base metaclass for class-nased command groups."""

    @classmethod
    def _construct_group(cls, name: str, description: str, **kwargs: Any) -> GT:
        """Construct a the actual group object."""
        raise NotImplementedError

    def __new__(
        cls,
        class_name: str,
        parents: tuple[type],
        attrs: dict[str, Any],
        name: Optional[str] = None,
        _base_class: bool = False,
        **kwargs: Any,
    ) -> Union[GT, type]:
        """Create a new command group class."""
        new_class = super().__new__(cls, class_name, parents, attrs)
        if _base_class:
            return new_class
        group = cls._construct_group(
            name=name or class_name_to_group_name(class_name),
            description=attrs["__doc__"] or "No description.",
            **kwargs,
        )
        # Command groups are singletons.
        instance = new_class()
        for attr in attrs.values():
            if isinstance(attr, SlashSubCommand):
                attr.prepend_params = (instance,)
            elif not isinstance(attr, ChildSlashCommand):
                continue
            group.subcommands[attr.name] = attr
        return group


class CommandGroupMeta(BaseCommandGroupMeta[SlashCommandGroup]):
    """Metaclass for class-based top-level command groups."""

    @classmethod
    def _construct_group(
        cls,
        name: str,
        description: str,
        guild_id: GuildID = GUILD_ID_DEFAULT,
        # This option only exists for legacy reasons, and has no effect.
        default_permission: bool | None = None,
    ) -> SlashCommandGroup:
        """Construct a the actual group object."""
        if default_permission is not None:
            warn_permissions_deprecation()
        return SlashCommandGroup(
            # The client will replace the guild ID default when the group is registered.
            guild_id=guild_id,  # type: ignore
            name=name,
            description=description,
        )


class CommandSubGroupMeta(BaseCommandGroupMeta[SlashCommandSubGroup]):
    """Metaclass for class-based command subgroups."""

    @classmethod
    def _construct_group(cls, name: str, description: str) -> SlashCommandSubGroup:
        """Construct a the actual group object."""
        return SlashCommandSubGroup(name=name, description=description)


class CommandGroup(metaclass=CommandGroupMeta, _base_class=True):
    """Class to inherit to build command groups.

    See `CommandClient.group` for more information and examples.
    """


class CommandSubGroup(metaclass=CommandSubGroupMeta, _base_class=True):
    """Class to inherit to build command subgroups."""


def subcommand(
    name: Optional[str] = None, description: Optional[str] = None
) -> SlashSubCommandConstructor:
    """Build a subcommand for within a group."""
    return SlashSubCommandConstructor(
        parent=None,
        name=name,
        description=description,
    )



"""Tools for setting permissions on commands."""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from .commands import TopLevelCommand

PermissionsSetter = Callable[["TopLevelCommand"], "TopLevelCommand"]


def warn_permissions_deprecation():
    """Warn about trying to configure permissions."""
    warnings.warn(
        "dslash: permissions configuration is deprecated and has no effect, as "
        "Discord no longer allows bots to configure permissions.",
        DeprecationWarning,
        stacklevel=3,
    )


def permissions_wrapper(*args: Any, **kwargs: Any) -> PermissionsSetter:
    """Make a no-op command wrapper."""
    warn_permissions_deprecation()
    return lambda command: command


global_permissions = guild_permissions = permissions_wrapper
allow_roles = disallow_roles = allow_users = disallow_users = permissions_wrapper



"""Interface for specifying option choices."""
from __future__ import annotations

from typing import Any, Generic, Optional, Type, TypeVar, Union

ChoiceType = Union[str, int, float]
CT = TypeVar("CT")


class Choice(Generic[CT]):
    """A single choice as part of an option's choices."""

    def __init__(self, name: str, value: CT):
        """Store the choice attributes."""
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        """Return the string representation of the choice."""
        return f"Choice(name={self.name!r}, value={self.value!r})"


class ChoicesMeta(type, Generic[CT]):
    """A metaclass for choices."""

    @classmethod
    def attribute_to_choice(
        cls, name: str, value: Any
    ) -> Optional[tuple[Choice[ChoiceType], Type[ChoiceType]]]:
        """Convert an attribute to a choice."""
        if isinstance(value, str):
            # 'name' is the name of the Python attribute, which should be the
            # internally-used *value* for the choice, and 'value' is the value
            # of the Python attribute, which should be the user-displayed
            # *name* of the choice.
            return Choice(name=value, value=name), str
        elif isinstance(value, Choice):
            return value, type(value.value)
        return None

    @classmethod
    def attributes_to_choices(
        cls,
        attrs: dict[str, Any],
    ) -> tuple[dict[str, Choice[CT]], Type[CT]]:
        """Get a mapping of attr name to chocie and the choice type for a choices class."""
        choices = {}
        choice_type = None
        for name, value in attrs.items():
            if name.startswith("_"):
                continue
            if not (choice_and_type := cls.attribute_to_choice(name, value)):
                continue
            choice, this_type = choice_and_type
            if choice_type is None:
                choice_type = this_type
            elif choice_type != this_type:
                raise TypeError(
                    f"All choices must be of the same type, but {name} is "
                    f"{this_type} and previous choices were {choice_type}."
                )
            choices[name] = choice
        if choice_type is None:
            raise ValueError("There must be at least one choice.")
        return choices, choice_type  # type: ignore

    def __new__(
        cls,
        class_name: str,
        parents: tuple[type],
        attrs: dict[str, Any],
        abstract: bool = False,
    ) -> type:
        """Create a new command group class."""
        if abstract:
            return super().__new__(cls, class_name, parents, attrs)
        attrs_to_choices, choice_type = cls.attributes_to_choices(attrs)
        for attr in attrs_to_choices:
            del attrs[attr]
        value_map = {}
        name_map = {}
        attrs["_value_map"] = value_map
        attrs["_name_map"] = name_map
        attrs["_choice_data"] = [
            {"name": choice.name, "value": choice.value} for choice in attrs_to_choices.values()
        ]
        attrs["_choice_type"] = choice_type
        new_cls = super().__new__(cls, class_name, parents, attrs)
        for name, choice_data in attrs_to_choices.items():
            choice = new_cls(choice_data.name, choice_data.value)
            value_map[choice_data.value] = choice
            name_map[name] = choice
        return new_cls

    def __getattr__(cls: Type[Choices[CT]], name: str) -> Choices[CT]:
        """Get a choice by name."""
        if name.startswith("__"):
            return super().__getattr__(name)
        return cls._name_map[name]  # type: ignore


class Choices(Generic[CT], metaclass=ChoicesMeta, abstract=True):
    """A class containing each of the choices for some option.

    Example usage:

    ```python
    class IcecreamFlavors(Choices):
        vanilla = "Vanilla"
        chocolate = "Chocolate"
        strawberry = "Strawberry"
    ```

    Or for integer options:

    ```python
    class Rating(Choices):
        one = Choice("Very bad", 1)
        two = Choice("Bad", 2)
        three = Choice("OK", 3)
        four = Choice("Good", 4)
        five = Choice("Great", 5)
    ```
    """

    # Discord choice value -> choice instance (class attribute)
    _value_map: dict[CT, Choice[CT]]

    # Python choice name -> choice instance (class attribute)
    _name_map: dict[str, Choice[CT]]

    # Discord JSON choice data (class attribute)
    _choice_data: dict[str, dict[str, Union[str, CT]]]

    # Choice type (class attribute)
    _choice_type: Type[CT]

    # Discord choice name (instance attribute)
    name: str

    # Discord choice value (instance attribute)
    value: CT

    @classmethod
    def _get_by_value(cls, value: CT) -> Choice[CT]:
        """Get a choice by value."""
        return cls._value_map[value]

    def __init__(self, name: str, value: CT):
        """Store the attributes of a single choice."""
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        """Return the string representation of the choice."""
        return f"{self.__class__.__name__}(name={self.name!r}, value={self.value!r})"



"""Models to represent arguments to slash commands."""
import types
import typing
from typing import Any, Optional, Type, Union

import nextcord
from nextcord.enums import Enum

from .choices import Choices

Channel = nextcord.abc.GuildChannel
Mentionable = Union[nextcord.User, nextcord.Member, nextcord.Role]


class ApplicationCommandOptionType(Enum):
    """The type of an argument to a slash command."""

    sub_command = 1
    sub_command_group = 2
    string = 3
    integer = 4
    boolean = 5
    user = 6
    channel = 7
    role = 8
    mentionable = 9
    number = 10
    attachment = 11


class CommandOption:
    """An argument to a slash command."""

    def __init__(self, *, name: str, description: str, type: Type):
        """Set up the command option."""
        self.description = description
        self.name = name
        self.type, self.choices, self.required = self._get_type_metadata(type)

    def _get_optional_type(self, param_type: Type) -> Optional[Type]:
        """Return the root type if 'type' is optional."""
        if typing.get_origin(param_type) in (typing.Union, types.UnionType):
            args = typing.get_args(param_type)
            if len(args) == 2 and isinstance(None, args[1]):
                return args[0]
        return None

    def _get_type_metadata(self, type: Type) -> tuple[Type, Optional[Type[Choices]], bool]:
        """Get the root type, associated choices, and whether it is required."""
        if root_type := self._get_optional_type(type):
            type = root_type
            required = False
        else:
            required = True
        if issubclass(type, Choices):
            choices = type
            type = choices._choice_type
        else:
            choices = None
        return type, choices, required

    def dump(self) -> dict[str, Any]:
        """Get the JSON data for registering this option with the API."""
        if self.choices:
            choice_dump = self.choices._choice_data
        else:
            choice_dump = None
        return {
            "name": self.name,
            "description": self.description,
            "required": self.required,
            "type": self.type_value.value,
            "choices": choice_dump,
        }

    @property
    def _is_user_type(self) -> bool:
        """Check if the option type is 'user'."""
        return (
            issubclass(self.type, (nextcord.User, nextcord.Member))
            or self.type is nextcord.abc.User
        )

    @property
    def _is_channel_type(self) -> bool:
        """Check if the option type is 'channel'."""
        return (
            issubclass(
                self.type,
                (nextcord.TextChannel, nextcord.VoiceChannel, nextcord.CategoryChannel),
            )
            or self.type is nextcord.abc.GuildChannel
        )

    @property
    def type_value(self) -> ApplicationCommandOptionType:
        """Get the Discord API code for the option's type."""
        if (not self.type) or issubclass(self.type, str):
            return ApplicationCommandOptionType.string
        if issubclass(self.type, bool):
            return ApplicationCommandOptionType.boolean
        if issubclass(self.type, int):
            return ApplicationCommandOptionType.integer
        if issubclass(self.type, float):
            return ApplicationCommandOptionType.number
        if self._is_user_type:
            return ApplicationCommandOptionType.user
        if self._is_channel_type:
            return ApplicationCommandOptionType.channel
        if issubclass(self.type, nextcord.Role):
            return ApplicationCommandOptionType.role
        if issubclass(self.type, nextcord.Attachment):
            return ApplicationCommandOptionType.attachment
        if typing.get_origin(self.type) is typing.Union:  # type: ignore
            if typing.get_args(self.type) == typing.get_args(Mentionable):
                return ApplicationCommandOptionType.mentionable
            raise TypeError("Union type not allowed for option type (except Mentionable).")
        raise TypeError(f"Unsupported option type {self.type!r}.")

    async def __call__(
        self, data: Optional[dict[str, Any]], interaction: nextcord.Interaction
    ) -> Any:
        """Process option data from the API."""
        if (not data) or "value" not in data:
            return None
        value = data["value"]
        type = ApplicationCommandOptionType(data["type"])
        if type in (
            ApplicationCommandOptionType.string,
            ApplicationCommandOptionType.boolean,
            ApplicationCommandOptionType.integer,
            ApplicationCommandOptionType.number,
        ):
            if self.choices:
                return self.choices._get_by_value(value)
            return value
        if type == ApplicationCommandOptionType.attachment:
            data = (
                interaction.data.get("resolved", {}).get("attachments", {}).get(value)
                if interaction.data
                else None
            )
            if data:
                return nextcord.Attachment(data=data, state=interaction._state)
            raise ValueError("Attachment option recieved without resolved attachment.")
        # We could use `resolved` for users/roles/channels, but working with
        # discord.py/nextcord internals is more hassle than it's worth.
        value = int(value)
        if not (guild := interaction.guild):
            raise ValueError("User/role/channel option recieved without a guild.")
        if type == ApplicationCommandOptionType.mentionable:
            if role := guild.get_role(value):
                return role
        if type in (
            ApplicationCommandOptionType.user,
            ApplicationCommandOptionType.mentionable,
        ):
            if not (user := guild.get_member(value)):
                user = await guild.fetch_member(value)
            return user
        if type == ApplicationCommandOptionType.role:
            return guild.get_role(value)
        if type == ApplicationCommandOptionType.channel:
            return guild.get_channel(value)



"""Discord client type."""
import logging
import re
import traceback
import typing
from collections import defaultdict
from typing import Any, Callable, Coroutine, Literal, Optional, Type, Union, overload

import nextcord
import nextcord.http
from nextcord.types.interactions import ApplicationCommand, EditApplicationCommand

from .commands import (
    Permissions,
    SlashCommandConstructor,
    SlashCommandGroup,
    SlashCommandInvokeError,
    TopLevelCommand,
)
from .permissions import warn_permissions_deprecation

if typing.TYPE_CHECKING:
    from .groups import CommandGroup

logger = logging.getLogger("dslash")
AsyncFunc = Callable[..., Coroutine]
AsyncWrapper = Callable[[AsyncFunc], AsyncFunc]

GUILD_ID_DEFAULT = "default"
GuildID = Union[None, int, Literal["default"]]


class CommandClient(nextcord.Client):
    """Client capable of registering and responding to slash commands.

    This supports all the same parameters, methods and attributes as a
    `nextcord.Client`, plus a few more documented below.
    """

    application_id: int
    _commands: dict[Optional[int], dict[str, TopLevelCommand]]

    def __init__(
        self,
        guild_id: Optional[int] = None,
        *,
        custom_interaction: Optional[Callable[[nextcord.Interaction], typing.Any]] = None,
        **options: typing.Any,
    ):
        """Set up the client.

        - `guild_id` (`Optional[int]`)

          A guild ID to set by default for all commands. If this is passed, it
          will be set as the guild ID for every command unless a guild ID is set
          for that command specifically. This is useful during testing, because
          guild-specific commands update immediately, whereas global commands
          can be cached for up to an hour.

        - `custom_interaction` (`Optional[Callable[[nextcord.Interaction], Any]]`)

          A callable which transforms an interaction into a custom interaction
          type to be passed as the first argument to commands.
        """
        super().__init__(**options)
        self.guild_id = guild_id
        self.custom_interaction = custom_interaction
        self._commands = defaultdict(dict)
        self._commands_by_id: dict[int, TopLevelCommand] = {}
        self._http: nextcord.http.HTTPClient = self._connection.http
        self._event_handlers: dict[str, list[AsyncFunc]] = defaultdict(list)

    def dispatch(self, event_name: str, *args: Any, **kwargs: Any) -> None:
        """Dispatch an event and call additional handlers."""
        super().dispatch(event_name, *args, **kwargs)
        for handler in self._event_handlers[event_name]:
            self._schedule_event(handler, event_name, *args, **kwargs)

    @overload
    def listener(self, value: AsyncFunc) -> AsyncFunc:
        """Register an event listener for the event with the function's name.

        The function should be named "on_" followed by the event name, for example:

        ```python
        @client.listener
        async def on_message(message):
            ...
        ```

        This would register a handler for the "message" event.
        """
        ...

    @overload
    def listener(self, value: str) -> AsyncWrapper:
        """Register an event listener with the given name.

        For example:

        ```python
        @client.listener("message")
        async def my_message_handler(message):
            ...
        ```
        """
        ...

    def listener(self, value: Union[str, AsyncFunc]) -> Union[AsyncFunc, AsyncWrapper]:
        """Register an event listener."""
        if isinstance(value, str):

            def wrapper(handler: AsyncFunc) -> AsyncFunc:
                self.add_listener(value, handler)
                return handler

            return wrapper
        elif callable(value):
            self.add_listener(value.__name__.removeprefix("on_"), value)
            return value
        else:
            raise TypeError(
                "CommandClient.listener should be passed an event name or event handler."
            )

    def add_listener(self, event_name: str, handler: AsyncFunc):
        """Register a function as an event listener."""
        self._event_handlers[event_name].append(handler)

    async def on_interaction(self, interaction: nextcord.Interaction):
        """Handle a slash command interaction being sent.

        If you override this method, or regiser an `on_interaction` listener
        with `@client.event`, you must make sure to call this method, or
        slash commands will not work.
        """
        if interaction.type == nextcord.InteractionType.application_command:
            await self.wait_until_ready()
            if (not interaction.data) or ("id" not in interaction.data):
                return
            command_id = int(interaction.data["id"])  # type: ignore
            try:
                await self._commands_by_id[command_id](interaction)
            except SlashCommandInvokeError as exc:
                self.dispatch("slash_command_error", interaction, exc)

    async def on_slash_command_error(
        self, interaction: nextcord.Interaction, error: SlashCommandInvokeError
    ):
        """Handle an error while invoking a slash command.

        You can override this method, or register an `on_slash_command_error`
        listener with `@client.event` to handle errors yourself.
        """
        # Log the error:
        original = error.original
        message = "".join(
            traceback.format_exception(type(original), original, original.__traceback__)
        )
        logger.error(f"An error occured while handling a command:\n{message}")
        # Send an error message:
        await interaction.response.send_message(
            embed=nextcord.Embed(title="Command error!", description=str(error), color=0xFF0000),
            ephemeral=True,
        )

    async def login(self, token: str):
        """Log in with a token and register all commands.

        If you run the client with `client.start`, you do not need to call this
        yourself, it will be called automatically.
        """
        data = await self.http.static_login(f"Bot {token.strip()}")
        if not self._connection.application_id:
            self._connection.application_id = int(data["id"])
        logger.info("Syncing commands...")
        self._commands[None] = self._commands.get(None, {})
        for scope, commands in self._commands.items():
            await self._update_scope_commands(scope, commands)
        logger.info("Finished syncing commands.")

    async def _update_scope_commands(
        self, scope: typing.Optional[int], commands: dict[str, TopLevelCommand]
    ):
        """Update all commands for a scope."""
        command_data = [command.dump() for command in commands.values()]
        try:
            created_commands = await self._register_scope_commands(
                scope, command_data  # type: ignore
            )
        except nextcord.HTTPException as error:
            self._handle_register_error(error, command_data)  # type: ignore
        for command_data in created_commands:
            command = commands[command_data["name"]]
            command_id = int(command_data["id"])
            self._commands_by_id[command_id] = command
            command.id = command_id

    async def _register_scope_commands(
        self, scope: typing.Optional[int], commands: list[EditApplicationCommand]
    ) -> list[ApplicationCommand]:
        """Register commands for a guild or globally."""
        if scope:
            logger.debug(f"Registering commands for guild {scope}.")
            return await self._http.bulk_upsert_guild_commands(
                self.application_id,
                scope,
                commands,
            )
        logger.debug("Registering global commands.")
        return await self._http.bulk_upsert_global_commands(
            self.application_id,
            commands,
        )

    def _handle_register_error(
        self, error: nextcord.HTTPException, commands: list[EditApplicationCommand]
    ):
        """Replace references to commands by index to references by name."""
        if error.status == 400:
            command_indices = set(re.findall(r"In\s(\d).", error.args[0]))
            error_string = error.args[0]
            for index in command_indices:
                error_command = commands[int(index)]
                error_string = error_string.replace(
                    f"In {index}",
                    "In {name}".format(**error_command),
                )
            error.args = (error_string, *error.args[1:])
        raise error

    def _store_command(self, command: TopLevelCommand):
        """Store a command to be registered on login."""
        self._commands[command.guild_id][command.name] = command

    def command(
        self,
        *,
        guild_id: GuildID = GUILD_ID_DEFAULT,
        default_permission: bool | None = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        permissions: Permissions = None,
    ) -> SlashCommandConstructor:
        """Register a new top-level slash command.

        - `guild_id` (`Optional[int]`)

          The ID of a guild to limit the command to.

        - `default_permission` (default `None`)

          THIS OPTION IS DEPRECATED AND HAS NO EFFECT. It remains for legacy reasons.

          Whether or not this command should be usable for people where no
          relevant permissions have been set.

        - `name` (`Optional[str]`)

          The name of the command. Defaults to the name of the function
          associated with it.

        - `description` (`Optional[str]`)

          A description of the command. Defaults to the first line of the
          docstring of the function associated with it.

        - `permissions`
          (`Optional[Union[PermissionsSetter, list[PermissionsSetter]]`)

          THIS OPTION IS DEPRECATED AND HAS NO EFFECT. It remains for legacy reasons.

          Either a single permission or a list of permissions (or `None`, the
          default). These can also be applied as decorators.

        This is intended to be used as a decorator. For example:

        ```python
        client = CommandClient()

        @client.command()
        async def flip(interaction: Interaction):
            \"""Flip a coin.\"""
            outcome = random.choice(['heads', 'tails'])
            await interaction.response.send_message(f'You got: {outcome}.')

        client.run(TOKEN)
        ```

        Example of setting more parameters:

        ```python
        @client.command(name='del')
        async def del_(
            interaction: Interaction,
            channel: Channel = option('The channel to delete.')
        ):
            \"""Delete a channel.\"""
            await channel.delete()
            await interaction.response.send_message(
                'Deleted that channel.',
                ephemeral=True
            )
        ```
        """
        if (permissions is not None) or (default_permission is not None):
            warn_permissions_deprecation()
        return SlashCommandConstructor(
            client=self,
            name=name,
            description=description,
            guild_id=self.guild_id if guild_id == GUILD_ID_DEFAULT else guild_id,
        )

    def group(self, group: Union[SlashCommandGroup, Type["CommandGroup"]]) -> SlashCommandGroup:
        """Register a top-level command group.

        Example usage:

        ```python
        client = CommandClient(guild_id=GUILD_ID)

        @client.group
        class Profiles(CommandGroup):
            \"""Commands to manage user profiles.\"""

            @subcommand()
            async def create(self, interaction: Interaction, user: User, age: int):
                \"""Create a new profile for someone.

                :user: The user to create the profile for.
                :age: The age of the user, in years.
                \"""
                ...

            @subcommand()
            async def delete(self, interaction: Interaction, user: User):
                \"""Delete someone's profile.

                :user: The person who's profile you want to delete.
                \"""
                ...

        client.run(TOKEN)
        ```
        """
        # The metaclass of SlashCommandGroup means that the name type checkers
        # think points to the type itself will in fact point to an instance of
        # SlashCommandGroup.
        group_: SlashCommandGroup = group  # type: ignore
        if group_.guild_id == GUILD_ID_DEFAULT:
            group_.guild_id = self.guild_id
        self._store_command(group_)
        return group_



"""A library to supplement Discord.py by adding support for slash commands."""
from .choices import Choice, Choices
from .client import CommandClient
from .commands import SlashCommandInvokeError
from .groups import CommandGroup, CommandSubGroup, subcommand
from .options import Channel, Mentionable
from .permissions import (
    allow_roles,
    allow_users,
    disallow_roles,
    disallow_users,
    global_permissions,
    guild_permissions,
)

__version__ = "0.7.1"
__all__ = (
    "__version__",
    "CommandClient",
    "CommandGroup",
    "CommandSubGroup",
    "subcommand",
    "Channel",
    "Mentionable",
    "Choices",
    "Choice",
    "allow_roles",
    "allow_users",
    "disallow_roles",
    "disallow_users",
    "guild_permissions",
    "global_permissions",
    "SlashCommandInvokeError",
)
